import {AfterViewInit, Component, ElementRef, OnInit, ViewChild} from '@angular/core';
import {ToasterService} from '../../core/toaster.service';
import {ActivatedRoute, Router} from '@angular/router';
import {AccountService} from '../../core/account.service';
import {Account} from '../../core/account';
import {AccountType} from '../../core/account-type';
import {CostCentre} from '../../core/cost-centre';
import {ConfirmDialogComponent} from '../../shared/confirm-dialog/confirm-dialog.component';
import { MatDialog } from '@angular/material/dialog';
import {FormBuilder, FormGroup} from '@angular/forms';

@Component({
  selector: 'app-account-detail',
  templateUrl: './account-detail.component.html',
  styleUrls: ['./account-detail.component.css']
})
export class AccountDetailComponent implements OnInit, AfterViewInit {
  @ViewChild('nameElement', { static: true }) nameElement: ElementRef;
  form: FormGroup;
  accountTypes: AccountType[];
  costCentres: CostCentre[];
  item: Account;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private fb: FormBuilder,
    private toaster: ToasterService,
    private ser: AccountService
  ) {
    this.createForm();
  }

  createForm() {
    this.form = this.fb.group({
      code: {value: '', disabled: true},
      name: '',
      type: '',
      isActive: '',
      isReconcilable: '',
      costCentre: ''
    });
  }

  ngOnInit() {
    this.route.data
      .subscribe((data: { item: Account, accountTypes: AccountType[], costCentres: CostCentre[] }) => {
        this.accountTypes = data.accountTypes;
        this.costCentres = data.costCentres;
        this.showItem(data.item);
      });
  }

  showItem(item: Account) {
    this.item = item;
    this.form.setValue({
      code: this.item.code || '(Auto)',
      name: this.item.name || '',
      type: this.item.type,
      isActive: this.item.isActive,
      isReconcilable: this.item.isReconcilable,
      costCentre: this.item.costCentre.id
    });
  }

  ngAfterViewInit() {
    setTimeout(() => {
      this.nameElement.nativeElement.focus();
    }, 0);
  }

  save() {
    this.ser.saveOrUpdate(this.getItem())
      .subscribe(
        (result) => {
          this.toaster.show('Success', '');
          this.router.navigateByUrl('/accounts');
        },
        (error) => {
          this.toaster.show('Danger', error);
        }
      );
  }

  delete() {
    this.ser.delete(this.item.id)
      .subscribe(
        (result) => {
          this.toaster.show('Success', '');
          this.router.navigateByUrl('/accounts');
        },
        (error) => {
          this.toaster.show('Danger', error);
        }
      );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: {title: 'Delete Account?', content: 'Are you sure? This cannot be undone.'}
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  getItem(): Account {
    const formModel = this.form.value;
    this.item.name = formModel.name;
    this.item.type = formModel.type;
    this.item.isActive = formModel.isActive;
    this.item.isReconcilable = formModel.isReconcilable;
    this.item.costCentre.id = formModel.costCentre;
    return this.item;
  }
}
