import {Injectable} from '@angular/core';
import {Observable} from 'rxjs/internal/Observable';
import {catchError} from 'rxjs/operators';
import {HttpClient, HttpHeaders, HttpParams} from '@angular/common/http';
import {Account} from './account';
import {ErrorLoggerService} from './error-logger.service';

const httpOptions = {
  headers: new HttpHeaders({'Content-Type': 'application/json'})
};

const url = '/api/accounts';
const serviceName = 'AccountService';

@Injectable({providedIn: 'root'})
export class AccountService {

  constructor(private http: HttpClient, private log: ErrorLoggerService) {
  }

  get(id: string): Observable<Account> {
    const getUrl: string = (id === null) ? `${url}` : `${url}/${id}`;
    return <Observable<Account>>this.http.get<Account>(getUrl)
      .pipe(
        catchError(this.log.handleError(serviceName, `get id=${id}`))
      );
  }

  list(): Observable<Account[]> {
    return <Observable<Account[]>>this.http.get<Account[]>(`${url}/list`)
      .pipe(
        catchError(this.log.handleError(serviceName, 'list'))
      );
  }

  paymentAutocomplete(query: string): Observable<Account[]> {
    const options = {params: new HttpParams().set('q', query).set('t', '1')};
    return <Observable<Account[]>>this.http.get<Account[]>(`${url}/query`, options)
      .pipe(
        catchError(this.log.handleError(serviceName, 'list'))
      );
  }

  receiptAutocomplete(query: string): Observable<Account[]> {
    const options = {params: new HttpParams().set('q', query).set('t', '1')};
    return <Observable<Account[]>>this.http.get<Account[]>(`${url}/query`, options)
      .pipe(
        catchError(this.log.handleError(serviceName, 'list'))
      );
  }

  save(account: Account): Observable<Account> {
    return <Observable<Account>>this.http.post<Account>(`${url}`, account, httpOptions)
      .pipe(
        catchError(this.log.handleError(serviceName, 'save'))
      );
  }

  update(account: Account): Observable<Account> {
    return <Observable<Account>>this.http.put<Account>(`${url}/${account.id}`, account, httpOptions)
      .pipe(
        catchError(this.log.handleError(serviceName, 'update'))
      );
  }

  saveOrUpdate(account: Account): Observable<Account> {
    if (!account.id) {
      return this.save(account);
    } else {
      return this.update(account);
    }
  }

  delete(id: string): Observable<Account> {
    return <Observable<Account>>this.http.delete<Account>(`${url}/${id}`, httpOptions)
      .pipe(
        catchError(this.log.handleError(serviceName, 'delete'))
      );
  }

  autocomplete(query: string): Observable<Account[]> {
    const options = {params: new HttpParams().set('q', query)};
    return <Observable<Account[]>>this.http.get<Account[]>(`${url}/query`, options)
      .pipe(
        catchError(this.log.handleError(serviceName, 'autocomplete'))
      );
  }

  balance(id: string, date: string): Observable<any> {
    const options = {params: new HttpParams().set('d', date)};
    return <Observable<any>>this.http.get<any>(`${url}/${id}/balance`, options)
      .pipe(
        catchError(this.log.handleError(serviceName, 'balance'))
      );
  }
}
