import {AfterViewInit, Component, ElementRef, OnInit, ViewChild} from '@angular/core';
import { MatPaginator } from '@angular/material/paginator';
import { MatSort } from '@angular/material/sort';
import {EmployeeListDataSource} from './employee-list-datasource';
import {Employee} from '../employee';
import {ActivatedRoute} from '@angular/router';
import {FormBuilder, FormGroup} from '@angular/forms';
import {Observable} from 'rxjs';
import {debounceTime, distinctUntilChanged, startWith} from 'rxjs/operators';
import {ToCsvService} from '../../shared/to-csv.service';

@Component({
  selector: 'app-employee-list',
  templateUrl: './employee-list.component.html',
  styleUrls: ['./employee-list.component.css']
})
export class EmployeeListComponent implements OnInit, AfterViewInit {
  @ViewChild('filterElement', { static: true }) filterElement: ElementRef;
  @ViewChild(MatPaginator, { static: true }) paginator: MatPaginator;
  @ViewChild(MatSort, { static: true }) sort: MatSort;
  dataSource: EmployeeListDataSource;
  filter: Observable<string>;
  form: FormGroup;
  list: Employee[];

  /** Columns displayed in the table. Columns IDs can be added, removed, or reordered. */
  displayedColumns = ['code', 'name', 'designation', 'salary', 'points', 'department', 'joiningDate', 'leavingDate'];

  constructor(private route: ActivatedRoute, private fb: FormBuilder, private toCsv: ToCsvService) {
    this.createForm();
    this.filter = this.listenToFilterChange();
  }

  createForm() {
    this.form = this.fb.group({
      filter: ''
    });
  }

  listenToFilterChange() {
    return this.form.get('filter').valueChanges
      .pipe(
        startWith(''),
        debounceTime(150),
        distinctUntilChanged()
      );
  }

  ngOnInit() {
    this.route.data
      .subscribe((data: { list: Employee[] }) => {
        this.list = data.list;
      });
    this.dataSource = new EmployeeListDataSource(this.paginator, this.sort, this.filter, this.list);
  }

  ngAfterViewInit() {
    setTimeout(() => {
      this.filterElement.nativeElement.focus();
    }, 0);
  }

  exportCsv() {
    const headers = {
      Code: 'code',
      Name: 'name',
      Designation: 'designation',
      Salary: 'salary',
      Points: 'points',
      Department: 'department',
      JoiningDate: 'joiningDate',
      LeavingDate: 'leavingDate'
    };

    const csvData = new Blob([this.toCsv.toCsv(headers, this.dataSource.data)], {type: 'text/csv;charset=utf-8;'});
    const link = document.createElement('a');
    link.href = window.URL.createObjectURL(csvData);
    link.setAttribute('download', 'employees.csv');
    document.body.appendChild(link);
    link.click();
    document.body.removeChild(link);
  }
}
