import { DataSource } from '@angular/cdk/collections';
import { MatPaginator } from '@angular/material/paginator';
import { MatSort } from '@angular/material/sort';
import { map, tap } from 'rxjs/operators';
import { merge, Observable, of as observableOf } from 'rxjs';
import { Product } from '../../core/product';


export class ProductListDataSource extends DataSource<Product> {
  private filterValue: string;

  constructor(private paginator: MatPaginator, private sort: MatSort, private filter: Observable<string>, public data: Product[]) {
    super();
    this.filter = filter.pipe(
      tap(x => this.filterValue = x)
    );
  }

  connect(): Observable<Product[]> {
    const dataMutations = [
      observableOf(this.data),
      this.filter,
      this.paginator.page,
      this.sort.sortChange
    ];

    return merge(...dataMutations).pipe(
      map((x: any) => {
        return this.getFilteredData([...this.data]);
      }),
      tap((x: Product[]) => this.paginator.length = x.length)
    ).pipe(
      map((x: any) => {
        return this.getPagedData(this.getSortedData(x));
      })
    );
  }

  disconnect() {
  }

  private getFilteredData(data: Product[]): Product[] {
    const filter = (this.filterValue === undefined) ? '' : this.filterValue;
    return filter.split(' ').reduce((p: Product[], c: string) => {
      return p.filter(x => {
          const productString = (
            x.code + ' ' + x.name + ' ' + x.units + ' ' + x.productGroup + (x.isPurchased ? ' purchased' : ' made')
            + (x.isSold ? 'sold' : 'used') + (x.isActive ? 'active' : 'deactive')
          ).toLowerCase();
          return productString.indexOf(c) !== -1;
        }
      );
    }, Object.assign([], data));
  }

  private getPagedData(data: Product[]) {
    const startIndex = this.paginator.pageIndex * this.paginator.pageSize;
    return data.splice(startIndex, this.paginator.pageSize);
  }

  private getSortedData(data: Product[]) {
    if (!this.sort.active || this.sort.direction === '') {
      return data;
    }

    return data.sort((a, b) => {
      const isAsc = this.sort.direction === 'asc';
      switch (this.sort.active) {
        case 'name':
          return compare(a.name, b.name, isAsc);
        case 'productGroup':
          return compare(a.productGroup, b.productGroup, isAsc);
        case 'id':
          return compare(+a.id, +b.id, isAsc);
        default:
          return 0;
      }
    });
  }
}

/** Simple sort comparator for example ID/Name columns (for client-side sorting). */
function compare(a, b, isAsc) {
  return (a < b ? -1 : 1) * (isAsc ? 1 : -1);
}
