import {AfterViewInit, Component, ElementRef, OnInit, OnDestroy, ViewChild} from '@angular/core';
import {FormBuilder, FormGroup} from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import {ActivatedRoute, Router} from '@angular/router';
import {BehaviorSubject, fromEvent, Observable, of, of as observableOf, zip} from 'rxjs';
import {PurchaseDataSource} from './purchase-datasource';
import {Account} from '../core/account';
import {VoucherService} from '../core/voucher.service';
import {AccountService} from '../core/account.service';
import {DbFile, Inventory, Voucher} from '../core/voucher';
import * as moment from 'moment';
import {AuthService} from '../auth/auth.service';
import {ConfirmDialogComponent} from '../shared/confirm-dialog/confirm-dialog.component';
import {ToasterService} from '../core/toaster.service';
import {debounceTime, distinctUntilChanged, map, startWith, switchMap} from 'rxjs/operators';
import {PurchaseDialogComponent} from './purchase-dialog.component';
import {ImageDialogComponent} from '../shared/image-dialog/image-dialog.component';
import {ProductService} from '../product/product.service';
import {Product} from '../core/product';
import {Hotkey, HotkeysService} from 'angular2-hotkeys';

@Component({
  selector: 'app-purchase',
  templateUrl: './purchase.component.html',
  styleUrls: ['./purchase.component.css']
})
export class PurchaseComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('accountElement', { static: true }) accountElement: ElementRef;
  @ViewChild('productElement', { static: true }) productElement: ElementRef;
  @ViewChild('dateElement', { static: true }) dateElement: ElementRef;
  public inventoryObservable = new BehaviorSubject<Inventory[]>([]);
  dataSource: PurchaseDataSource;
  form: FormGroup;
  voucher: Voucher;
  product: Product;
  accBal: any;

  displayedColumns = ['product', 'quantity', 'rate', 'tax', 'discount', 'amount', 'action'];

  accounts: Observable<Account[]>;
  products: Observable<Product[]>;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private fb: FormBuilder,
    private dialog: MatDialog,
    private hotkeys: HotkeysService,
    private toaster: ToasterService,
    private auth: AuthService,
    private ser: VoucherService,
    private productSer: ProductService,
    private accountSer: AccountService
  ) {
    this.createForm();
    this.listenToAccountAutocompleteChange();
    this.listenToProductAutocompleteChange();
  }

  ngOnInit() {
    this.route.data
      .subscribe((data: { voucher: Voucher }) => {
        this.loadVoucher(data.voucher);
      });
    this.hotkeys.add(new Hotkey('f2', (event: KeyboardEvent): boolean => {
      setTimeout(() => {
        this.dateElement.nativeElement.focus();
      }, 0);
      return false; // Prevent bubbling
    }, ['INPUT', 'SELECT', 'TEXTAREA']));
    this.hotkeys.add(new Hotkey('ctrl+s', (event: KeyboardEvent): boolean => {
      if (this.canSave()) {
        this.save();
      }
      return false; // Prevent bubbling
    }, ['INPUT', 'SELECT', 'TEXTAREA']));
    this.hotkeys.add(new Hotkey('ctrl+p', (event: KeyboardEvent): boolean => {
      if (this.voucher.id && !this.voucher.posted && this.auth.user.perms.indexOf('post-vouchers') !== -1) {
        this.post();
      }
      return false; // Prevent bubbling
    }, ['INPUT', 'SELECT', 'TEXTAREA']));
  }

  ngAfterViewInit() {
    this.focusAccount();
  }

  ngOnDestroy() {
    this.hotkeys.reset();
  }

  loadVoucher(voucher) {
    this.voucher = voucher;
    this.form.setValue({
      date: moment(this.voucher.date, 'DD-MMM-YYYY').toDate(),
      account: this.voucher.vendor,
      amount: Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)),
      addRow: {
        product: '',
        quantity: '',
        price: '',
        tax: '',
        discount: ''
      },
      narration: this.voucher.narration
    });
    this.dataSource = new PurchaseDataSource(this.inventoryObservable);
    this.updateView();
  }

  focusAccount() {
    setTimeout(() => {
      this.accountElement.nativeElement.focus();
    }, 0);
  }

  addRow() {
    const formValue = this.form.get('addRow').value;
    const quantity = +(formValue.quantity);
    const price = +(formValue.price);
    const tax = +(formValue.tax);
    const discount = +(formValue.discount);
    if (this.product === null || quantity <= 0 || price <= 0) {
      return;
    }
    const oldFiltered = this.voucher.inventories.filter((x) => x.product.id === this.product.id);
    if (oldFiltered.length) {
      this.toaster.show('Danger', 'Product already added');
      return;
    }
    this.voucher.inventories.push({
      id: null,
      quantity: quantity,
      rate: price,
      tax: tax,
      discount: discount,
      amount: quantity * price * (1 + tax) * (1 - discount),
      product: this.product,
      batch: null
    });
    this.resetAddRow();
    this.updateView();
  }

  resetAddRow() {
    this.form.get('addRow').reset({
      product: null,
      quantity: '',
      price: '',
      tax: '',
      discount: ''
    });
    this.product = null;
    setTimeout(() => {
      this.productElement.nativeElement.focus();
    }, 0);
  }

  updateView() {
    this.inventoryObservable.next(this.voucher.inventories);
    this.form.get('amount').setValue(Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)));
  }

  editRow(row: Inventory) {
    const dialogRef = this.dialog.open(PurchaseDialogComponent, {
      width: '750px',
      data: {inventory: Object.assign({}, row)}
    });

    dialogRef.afterClosed().subscribe((result: boolean | Inventory) => {
      if (!result) {
        return;
      }
      const j = result as Inventory;
      if (j.product.id !== row.product.id && this.voucher.inventories.filter((x) => x.product.id === j.product.id).length) {
        return;
      }
      Object.assign(row, j);
      this.updateView();
    });
  }

  deleteRow(row: Inventory) {
    this.voucher.inventories.splice(this.voucher.inventories.indexOf(row), 1);
    this.updateView();
  }

  createForm() {
    this.form = this.fb.group({
      date: '',
      account: '',
      amount: {value: '', disabled: true},
      addRow: this.fb.group({
        product: '',
        quantity: '',
        price: '',
        tax: '',
        discount: ''
      }),
      narration: ''
    });
    this.accBal = null;
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    } else if (this.voucher.posted && this.auth.user.perms.indexOf('edit-posted-vouchers') !== -1) {
      return true;
    } else {
      return this.voucher.user.id === this.auth.user.id || this.auth.user.perms.indexOf("edit-other-user's-vouchers") !== -1;
    }
  }

  post() {
    this.ser.post(this.voucher.id)
      .subscribe(
        (result) => {
          this.loadVoucher(result);
          this.toaster.show('Success', 'Voucher Posted');
        },
        (error) => {
          this.toaster.show('Danger', error);
        }
      );
  }

  save() {
    this.ser.saveOrUpdate(this.getVoucher())
      .subscribe(
        (result) => {
          this.loadVoucher(result);
          this.toaster.show('Success', '');
          this.router.navigate(['/purchase', result.id]);
        },
        (error) => {
          this.toaster.show('Danger', error);
        }
      );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    this.voucher.vendor = formModel.account;
    this.voucher.narration = formModel.narration;
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id)
      .subscribe(
        (result) => {
          this.toaster.show('Success', '');
          this.router.navigate(['/purchase'], {replaceUrl: true});
        },
        (error) => {
          this.toaster.show('Danger', error);
        }
      );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: {title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.'}
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  listenToAccountAutocompleteChange(): void {
    const control = this.form.get('account');
    this.accounts = control.valueChanges
      .pipe(
        startWith(null),
        map(x => (x !== null && x.length >= 1) ? x : null),
        debounceTime(150),
        distinctUntilChanged(),
        switchMap(x => (x === null) ? observableOf([]) : this.accountSer.autocomplete(x))
      );
  }

  listenToProductAutocompleteChange(): void {
    const control = this.form.get('addRow').get('product');
    this.products = control.valueChanges
      .pipe(
        startWith(null),
        map(x => (x !== null && x.length >= 1) ? x : null),
        debounceTime(150),
        distinctUntilChanged(),
        switchMap(x => (x === null) ? observableOf([]) : this.productSer.autocomplete(x))
      );
  }

  displayAccount(account?: Account): string | undefined {
    return account ? account.name : undefined;
  }

  accountSelected(event: MatAutocompleteSelectedEvent): void {
    this.form.get('account').setValue(event.option.value);
  }

  displayProductName(product?: Product): string | undefined {
    return product ? product.name : undefined;
  }

  productSelected(event: MatAutocompleteSelectedEvent): void {
    this.product = event.option.value;
    this.form.get('addRow').get('price').setValue(this.product.price);
  }

  zoomImage(file: DbFile) {
    this.dialog.open(ImageDialogComponent, {
      width: '750px',
      data: file.resized
    });
  }

  deleteImage(file: DbFile) {
    const index = this.voucher.files.indexOf(file);
    this.voucher.files.splice(index, 1);
  }

  detectFiles(event) {
    const files = event.target.files;
    if (files) {
      for (const file of files) {
        const reader = new FileReader();
        reader.onload = (e: any) => {
          zip(of(e.target.result),
            this.resizeImage(e.target.result, 100, 150),
            this.resizeImage(e.target.result, 825, 1170)
          ).subscribe(
            val => this.voucher.files.push({id: null, thumbnail: val[1], resized: val[2]})
          );
        };
        reader.readAsDataURL(file);
      }
    }
  }

  resizeImage(image, MAX_WIDTH, MAX_HEIGHT) {
    const canvas = document.createElement('canvas');
    const ctx = canvas.getContext('2d');
    const img = new Image();
    const ex = fromEvent(img, 'load').pipe(
      map((e) => {
        let width = img.naturalWidth,
          height = img.naturalHeight;
        const ratio = Math.min(1, MAX_WIDTH / width, MAX_HEIGHT / height);

        if (ratio === 1) {
          return image;
        }

        width *= ratio;
        height *= ratio;

        canvas.width = width;
        canvas.height = height;
        ctx.drawImage(img, 0, 0, width, height);
        return canvas.toDataURL('image/jpeg', 0.95);
      })
    );
    img.src = image;
    return ex;
  }
}
