import {Component, OnInit, ViewChild} from '@angular/core';
import {FormBuilder, FormGroup} from '@angular/forms';
import { MatPaginator } from '@angular/material/paginator';
import { MatSort } from '@angular/material/sort';
import {ActivatedRoute, Router} from '@angular/router';
import {Observable} from 'rxjs';
import * as moment from 'moment';
import {RawMaterialCostDataSource} from './raw-material-cost-datasource';
import {RawMaterialCost} from './raw-material-cost';
import {Product} from '../core/product';
import {ToCsvService} from '../shared/to-csv.service';

@Component({
  selector: 'app-raw-material-cost',
  templateUrl: './raw-material-cost.component.html',
  styleUrls: ['./raw-material-cost.component.css']
})
export class RawMaterialCostComponent implements OnInit {
  @ViewChild(MatPaginator, { static: true }) paginator: MatPaginator;
  @ViewChild(MatSort, { static: true }) sort: MatSort;
  dataSource: RawMaterialCostDataSource;
  form: FormGroup;
  info: RawMaterialCost;
  debitQuantity: number;
  debitAmount: number;
  creditQuantity: number;
  creditAmount: number;
  runningQuantity: number;
  runningAmount: number;
  /** Columns displayed in the table. Columns IDs can be added, removed, or reordered. */
  displayedColumns: string[];
  columnsNoId = ['name', 'issue', 'sale', 'rmc'];
  columnsId = ['name', 'group', 'quantity', 'net', 'gross'];

  products: Observable<Product[]>;

  constructor(private route: ActivatedRoute, private router: Router, private fb: FormBuilder, private toCsv: ToCsvService) {
    this.createForm();
  }

  ngOnInit() {
    this.route.data
      .subscribe((data: { info: RawMaterialCost }) => {
        this.info = data.info;
        this.displayedColumns = (this.info.id) ? this.columnsId : this.columnsNoId;
        this.form.setValue({
          startDate: moment(this.info.startDate, 'DD-MMM-YYYY').toDate(),
          finishDate: moment(this.info.finishDate, 'DD-MMM-YYYY').toDate()
        });
        this.dataSource = new RawMaterialCostDataSource(this.paginator, this.sort, this.info.body);
      });
  }

  show() {
    const model = this.prepareSubmit();
    this.router.navigate(['raw-material-cost'], {
      queryParams: {
        startDate: model.startDate,
        finishDate: model.finishDate
      }
    });
  }

  createForm() {
    this.form = this.fb.group({
      startDate: '',
      finishDate: '',
    });
  }

  prepareSubmit(): RawMaterialCost {
    const formModel = this.form.value;

    return {
      startDate: moment(formModel.startDate).format('DD-MMM-YYYY'),
      finishDate: moment(formModel.finishDate).format('DD-MMM-YYYY'),
    };
  }

  exportCsv() {
    const headers = (this.info.id) ? {
      Name: 'name',
      Group: 'group',
      Quantity: 'quantity',
      Net: 'net',
      Gross: 'gross'
    } : {
      Name: 'name',
      Issue: 'issue',
      Sale: 'sale',
      RMC: 'rmc'
    };

    const csvData = new Blob([this.toCsv.toCsv(headers, this.dataSource.data)], {type: 'text/csv;charset=utf-8;'});
    const link = document.createElement('a');
    link.href = window.URL.createObjectURL(csvData);
    link.setAttribute('download', 'raw-material-cost.csv');
    document.body.appendChild(link);
    link.click();
    document.body.removeChild(link);
  }
}
